/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins;

import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.util.List;
import java.util.Vector;

import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.exceptions.NotSerialisedEntityException;


/**
 * Subclass of {@link Entity} for all entities which are entirely canonical
 * and therefore not updated while building state. I'm not totally certain
 * that we really need this, but it can't hurt and it helps make things a bit
 * clearer.
 *
 * @author Matthew Johnson <autoumpire@matthew.ath.cx>
 */
public abstract class CanonicalEntity extends Entity
{
   /** Set if the object has been modified */
   private boolean dirty = false;
   /** sets the dirty flag. To be called by subclasses on set methods */
   protected void setDirty() { dirty = true; }
   /** checks the dirty flag. This is checked by other classes in vanilla */
   boolean dirty() { return dirty; }
   /** resets the dirty flag. 
    * Should only be called by the database backend after writing the data */
   protected void clean() { dirty = false; }
   
   /** 
    * Takes a String serialised form and returns a CanonicalEntity
    * 
    * @param serial The serialised form
    * @return The corresponding CanonicalEntity
    */
   static CanonicalEntity deSerialise(String serial) throws NotSerialisedEntityException
   {
      // form to parse "__::full.class.name(int)"
      
      // check its a serialised object
      if (!serial.startsWith("__::")) throw new NotSerialisedEntityException();

      String entity = "";
      String id_str = "";
      Integer id = null;
      // parse details
      int i = 4;
      try {
         for (char c = serial.charAt(i); '(' != c; c = serial.charAt(++i))
            entity += c;
         i++;
         for (char c = serial.charAt(i); ')' != c; c = serial.charAt(++i))
            id_str += c;
         id = new Integer(id_str);
      } catch (ArrayIndexOutOfBoundsException AIOOBe) { if (Debug.debug) Debug.print(CanonicalEntity.class, "Exception: "+AIOOBe); throw new NotSerialisedEntityException(AIOOBe); 
      } catch (NumberFormatException NFe) { if (Debug.debug) Debug.print(CanonicalEntity.class, "Exception: "+NFe); throw new NotSerialisedEntityException(NFe); }
      // use reflection to instantiate class
      CanonicalEntity ce = null;
      try {
         Class c = Class.forName(entity);
         Method m = c.getDeclaredMethod("getByID", new Class[] {Integer.class});
         ce = (CanonicalEntity) m.invoke(null, new Object[] {id});
      } 
         
      // re-throw all the exceptions, wrapped
      catch (ClassNotFoundException CNFe) { if (Debug.debug) Debug.print(CanonicalEntity.class, "Exception: "+CNFe); throw new NotSerialisedEntityException(CNFe); } 
      catch (ClassCastException CCe) { if (Debug.debug) Debug.print(CanonicalEntity.class, "Exception: "+CCe); throw new NotSerialisedEntityException(CCe); } 
      catch (NoSuchMethodException NSMe) { if (Debug.debug) Debug.print(CanonicalEntity.class, "Exception: "+NSMe); throw new NotSerialisedEntityException(NSMe); } 
      catch (SecurityException Se) { if (Debug.debug) Debug.print(CanonicalEntity.class, "Exception: "+Se); throw new NotSerialisedEntityException(Se); } 
      // catch (InstantiationException Ie) { if (Debug.debug) Debug.print(CanonicalEntity.class, "Exception: "+Ie); throw new NotSerialisedEntityException(Ie); }
      catch (IllegalAccessException IAce) { if (Debug.debug) Debug.print(CanonicalEntity.class, "Exception: "+IAce); throw new NotSerialisedEntityException(IAce); }
      catch (IllegalArgumentException IAre) { if (Debug.debug) Debug.print(CanonicalEntity.class, "Exception: "+IAre); throw new NotSerialisedEntityException(IAre); }
      catch (InvocationTargetException ITe) { if (Debug.debug) Debug.print(CanonicalEntity.class, "Exception: "+ITe); throw new NotSerialisedEntityException(ITe); }
      
      // return it
      return ce;
   }

   static List getAll(CanonicalEntity ce) throws AssassinSQLException
   {
      if (ce instanceof Event)
         return Event.getAllEvents();
      else if (ce instanceof Assassin)
         return Assassin.getAllAssassinsOrderByRealName();
      else if (ce instanceof Player)
         return Player.getAllPlayersOrderByName();
      else if (ce instanceof Report)
         return Report.getAllReports();
      else if (ce instanceof StateChange)
         return StateChange.getAll();
      else if (ce instanceof Pseudonym)
         return Pseudonym.getAllPseudonyms();
      else
         return new Vector();
   }

   /** Override to return the class and database reference field value.
    * Format: "__::full.class.name(id)"
    */ 
   public abstract String toString();
   /** Override to return a printable string for the entity */
   public abstract String getPrintable() throws AssassinSQLException;
   
   abstract void readDetails() throws AssassinSQLException;
   abstract void writeDetails() throws AssassinSQLException;
}


