/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins;

import java.sql.ResultSet;
import java.sql.SQLException;
//import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.exceptions.NoDatabaseAssociativityException;
import org.ucam.srcf.assassins.exceptions.NonExistentEntityException;
import org.ucam.srcf.assassins.exceptions.ObjectModifiedException;


/**
 * Class representing events.
 *
 * @author Matthew Johnson <autoumpire@matthew.ath.cx>
 */
public class Event extends CanonicalEntity
{
   public static final int READALL = 0;
   public static final int READBUILTIN = 1;
   public static final int READSTATECHANGES = 2;
   public static final int READREPORTS = 3;
   public static final int READPARTICIPANTS = 4;
   private static Map eventCache = new HashMap();
   private static List newEventCache = new Vector();
   
   // Static methods:
   
   private static List getEventsFromQuery(String query) throws AssassinSQLException
   {
      ResultSet r = DatabaseBackend.query(query);
      List events = new Vector();

      try {
         while (r.next()) {
            int eid = r.getInt("eid");
            Integer EID = new Integer(eid);

            // check the cache and add as necessary
            if (eventCache.containsKey(EID)) 
               events.add(eventCache.get(EID));
            else {
               Event e = new Event(eid);
               e.headline = r.getString("headline");
               e.eventtime = r.getLong("time");
               e.modifytime = r.getLong("modified");
               e.active = new Boolean(r.getBoolean("active"));
               e.visible = new Boolean(r.getBoolean("visible"));
               eventCache.put(EID, e);            
               events.add(e);
            }
         }
         r.close();
      } catch (SQLException SQLe) { throw new AssassinSQLException(SQLe); }
      return events;
   }
    
   private static List getEventViewersFromQuery(String query) throws AssassinSQLException
   {
      ResultSet r = DatabaseBackend.query(query);
      List events = new Vector();

      try {
         while (r.next()) {
            int eid = r.getInt(1);
            Integer EID = new Integer(eid);

            // check the cache and add as necessary
            if (eventCache.containsKey(EID)) 
               events.add(new EventViewer((Event) eventCache.get(EID)));
            else {
               Event e = new Event(eid);
               e.headline = r.getString("headline");
               e.eventtime = r.getLong("time");
               e.modifytime = r.getLong("modified");
               e.active = new Boolean(r.getBoolean("active"));
               e.visible = new Boolean(r.getBoolean("visible"));
               eventCache.put(EID, e);            
               events.add(new EventViewer(e));
            }
         }
         r.close();
      } catch (SQLException SQLe) { throw new AssassinSQLException(SQLe); }
      return events;
   }

   /**
    * Precaches reports, statechanges and participants for all events
    *
    * @param from Time from
    * @param to Time until
    */
   static void preCacheSubClasses(long from, long to) throws AssassinSQLException
   {
      Iterator i = eventCache.values().iterator();
      while (i.hasNext()) {
         Event e = (Event) i.next();
         if (null == e.participants) e.participants = new HashMap();
         if (null == e.participantPseudonyms) e.participantPseudonyms = new HashMap();
         if (null == e.reports) e.reports = new Vector();
         if (null == e.stateChanges) e.stateChanges = new Vector();
      }
      try {
         // participants
         Pseudonym.preCache();
         Assassin.getAllAssassins();
         String query = "SELECT aid, evp.eid, psid, pindex FROM " +DatabaseBackend.getPrefix()+"_eventparticipants evp NATURAL JOIN " +DatabaseBackend.getPrefix()+"_events ev WHERE ev.time >= '"+from+"' AND ev.time <= '"+to+"'";
         ResultSet r = DatabaseBackend.query(query);
         while (r.next()) {
            Event e = (Event) Event.getByID(r.getInt("eid"));
            Pseudonym p = Pseudonym.getByPsid(r.getInt("psid"));
            Assassin a = Assassin.getByAid(r.getInt("aid"));
            Integer pindex = new Integer(r.getInt("pindex"));
            e.participants.put(pindex, a);
            e.participantPseudonyms.put(pindex, p);
         }
         r.close();

         // reports
         Report.preCache();
         query = "SELECT rid, event FROM "+DatabaseBackend.getPrefix()+"_reports re INNER JOIN " +DatabaseBackend.getPrefix()+"_events ev ON re.event=ev.eid WHERE ev.time >= '"+from+"' AND ev.time <= '"+to+"'";
         r = DatabaseBackend.query(query);
         while (r.next()) {
            Event e = (Event) Event.getByID(r.getInt("event"));
            Report rep = Report.getByRid(r.getInt("rid"));
            e.reports.add(rep);
         }
         r.close();

         // state changes
         StateChange.preCache();
         query = "SELECT scid, event FROM "+DatabaseBackend.getPrefix()+"_statechanges sc INNER JOIN " +DatabaseBackend.getPrefix()+"_events ev ON ev.eid=sc.event WHERE ev.time >= '"+from+"' AND ev.time <= '"+to+"'";
         r = DatabaseBackend.query(query);
         while (r.next()) {
            Event e = (Event) Event.getByID(r.getInt("event"));
            StateChange s = StateChange.getByScid(r.getInt("scid"));
            e.stateChanges.add(s);
         }
         r.close();
      } catch (SQLException SQLe) { throw new AssassinSQLException(SQLe); }
   }
     
    /**
    * Returns all the events with a headline
    *
    * @return A List of Events
    * @throws AssassinSQLException if there was an SQL problem
    */
   static List getAllEventsWithHeadline() throws AssassinSQLException
   {
      String query = "SELECT * FROM " + DatabaseBackend.getPrefix() +"_events WHERE headline != '' ORDER BY time";
      return getEventsFromQuery(query);
   }
  
    /**
    * Returns all the events
    *
    * @return A List of Events
    * @throws AssassinSQLException if there was an SQL problem
    */
   static List getAllEvents() throws AssassinSQLException
   {
      String query = "SELECT * FROM " + DatabaseBackend.getPrefix() +"_events ORDER BY time";
      return getEventsFromQuery(query);
   }
   
    /**
    * Returns all the events with a headline, ordered by most recently modified
    *
    * @return A List of Events
    * @throws AssassinSQLException if there was an SQL problem
    */
   static List getAllEventsWithHeadlineOrderByRecentlyModified() throws AssassinSQLException
   {
      String query = "SELECT * FROM " + DatabaseBackend.getPrefix() +"_events WHERE headline != '' "+
                     "ORDER BY modified DESC" ;
      return getEventsFromQuery(query);
   }
  
   /**
    * Returns all the events happenning at a particular time
    *
    * @param time The time to look for
    * @return A List of Events
    * @throws AssassinSQLException if there was an SQL problem
    */
   static List getEventsAtTime(long time) throws AssassinSQLException
   {
      String query = "SELECT * FROM " + DatabaseBackend.getPrefix() +"_events WHERE time = '"+time+"' ORDER BY time";
      return getEventsFromQuery(query);
   }

   /**
    * Returns all the events happenning in a particular range of times
    *
    * @param from The lower bound of the time (inclusive)
    * @param to The upper bound of the time (inclusive)
    * @return A List of Events
    * @throws AssassinSQLException if there was an SQL problem
    */
   static List getEventsInTime(long from, long to) throws AssassinSQLException
   {
      String query = "SELECT * FROM " + DatabaseBackend.getPrefix() +"_events WHERE time >= '"+from+"' AND time <= '"+to+"' ORDER BY time";
      return getEventsFromQuery(query);
   }

   /**
    * Returns all the events involving a particular assassin
    *
    * @param a The assassin to search for
    * @return A List of Events
    * @throws AssassinSQLException if there was an SQL problem
    */
   static List getEventsInvolving(Assassin a) throws AssassinSQLException
   {
      String query = "SELECT * FROM " + DatabaseBackend.getPrefix() +"_eventparticipants WHERE aid = '"+a.getAid()+"' ORDER BY time";
      return getEventsFromQuery(query);
   }
   
   /**
    * Returns all the events involving a particular assassin
    *
    * @param a The assassin to search for
    * @return A List of Events
    * @throws AssassinSQLException if there was an SQL problem
    */
   static List getEventsInvolving(AssassinViewer a) throws AssassinSQLException
   {
      String query = "SELECT * FROM " + DatabaseBackend.getPrefix() +"_eventparticipants WHERE aid = '"+((Assassin)a.getEntity()).getAid()+"'";
      return getEventsFromQuery(query);
   }

   /**
    * Returns all the events with a particular type of StateChange
    *
    * @param s The StateChange Field
    * @return A List of Events
    * @throws AssassinSQLException if there was an SQL problem
    */
   static List getEventsWithStateChangeField(String s) throws AssassinSQLException
   {
      String query = "SELECT event FROM " + DatabaseBackend.getPrefix() +"_statechanges WHERE field = '"+s+"'";
      return getEventsFromQuery(query);
   }
     
    /**
    * Returns all the events
    *
    * @return A List of EventViewers
    * @throws AssassinSQLException if there was an SQL problem
    */
   public static List getAllEventViewers() throws AssassinSQLException
   {
      String query = "SELECT * FROM " + DatabaseBackend.getPrefix() +"_events ORDER BY time";
      return getEventViewersFromQuery(query);
   }
     
    /**
    * Returns all the events with a headline
    *
    * @return A List of EventViewers
    * @throws AssassinSQLException if there was an SQL problem
    */
   public static List getAllEventViewersWithHeadline() throws AssassinSQLException
   {
      String query = "SELECT * FROM " + DatabaseBackend.getPrefix() +"_events where headline != '' ORDER BY time";
      return getEventViewersFromQuery(query);
   }
  
   /**
    * Returns all the events happenning at a particular time
    *
    * @param time The time to look for
    * @return A List of EventViewerss
    * @throws AssassinSQLException if there was an SQL problem
    */
   public static List getEventViewersAtTime(long time) throws AssassinSQLException
   {
      String query = "SELECT * FROM " + DatabaseBackend.getPrefix() +"_events WHERE time = '"+time+"' ORDER BY time";
      return getEventViewersFromQuery(query);
   }

   /**
    * Returns all the events happenning in a particular range of times
    *
    * @param from The lower bound of the time (inclusive)
    * @param to The upper bound of the time (inclusive)
    * @return A List of EventViewers
    * @throws AssassinSQLException if there was an SQL problem
    */
   public static List getEventViewersInTime(long from, long to) throws AssassinSQLException
   {
      String query = "SELECT * FROM " + DatabaseBackend.getPrefix() +"_events WHERE time >= '"+from+"' AND time <= '"+to+"' ORDER BY time";
      return getEventViewersFromQuery(query);
   }

   /**
    * Returns all the events involving a particular assassin
    *
    * @param a The assassin to search for
    * @return A List of EventViewers
    * @throws AssassinSQLException if there was an SQL problem
    */
   public static List getEventViewersInvolving(Assassin a) throws AssassinSQLException
   {
      String query = "SELECT * FROM " + DatabaseBackend.getPrefix() +"_eventparticipants WHERE aid = '"+a.getAid()+"'";
      return getEventsFromQuery(query);
   }
   
   /**
    * Returns all the events involving a particular assassin
    *
    * @param a The assassin to search for
    * @return A List of EventViewers
    * @throws AssassinSQLException if there was an SQL problem
    */
   public static List getEventViewersInvolving(AssassinViewer a) throws AssassinSQLException
   {
      String query = "SELECT * FROM " + DatabaseBackend.getPrefix() +"_eventparticipants WHERE aid = '"+((Assassin)a.getEntity()).getAid()+"'";
      return getEventsFromQuery(query);
   }

   /**
    * Returns all the events with a particular type of StateChange
    *
    * @param s The StateChange Field
    * @return A List of EventViewers
    * @throws AssassinSQLException if there was an SQL problem
    */
   public static List getEventViewersWithStateChangeField(String s) throws AssassinSQLException
   {
      String query = "SELECT event FROM " + DatabaseBackend.getPrefix() +"_statechanges WHERE field = '"+s+"'";
      return getEventViewersFromQuery(query);
   }

   static Event getByID(int eid) throws AssassinSQLException
   {
      if (0 == eid) return null;

      Event p;

      // check cache
      if (null == (p = (Event) eventCache.get(new Integer(eid)))) {
         // check database existence
         // String query = "SELECT eid FROM " + DatabaseBackend.getPrefix() + "_events WHERE eid = '"+eid+"'";
         // ResultSet r = DatabaseBackend.query(query);
         // try {
         //    if (!r.next()) throw new NonExistentEntityException();
         //    r.close();
         // } catch (SQLException SQLe) { throw new AssassinSQLException(SQLe); }

         // add to cache
         p = new Event(eid);
         eventCache.put(new Integer(eid), p);
      }
      return p;

   }
   
   static Event getByID(Integer eid) throws AssassinSQLException
   {
      return getByID(eid.intValue());
   }
   
   /**
    * Checks all Event objects for updates, and writes them out to the database.
    */
   static void commitAll() throws AssassinSQLException
   {
      // iterate over all new Event objects
      while (newEventCache.size() > 0) {
         Event a = (Event) newEventCache.remove(0);
         a.writeDetails();
         eventCache.put(new Integer(a.getEid()),a);
      }
      
      // iterate over all cached Event objects
      Iterator i = eventCache.values().iterator();
      while (i.hasNext()) {
         Event e = (Event) i.next();
         // if they are modified, then update
         if (e.dirty()) e.writeDetails();
      }
   }

   /**
    * Removes an event from the database and the pending list.
    *
    * Be Careful with this! Do not reuse a reference if you have deleted it!!
    */
   static void deleteEvent(Event e) throws AssassinSQLException
   {
      newEventCache.remove(e);
      eventCache.remove(e);
      Iterator i = e.getStateChanges().iterator();
      while (i.hasNext())
         ((StateChange) i.next()).setEvent(null);
      if (0 != e.eid) {
         String query = "DELETE FROM "+DatabaseBackend.getPrefix()+"_events WHERE eid = '"+e.eid+"'";
         DatabaseBackend.query(query);
         e.eid = 0;
      }
   }

   
   /** Database event ID */
   private int eid;
   private String headline = null;
   private long eventtime = 0;
   private long modifytime = 0;
   private Boolean visible = null;
   private Boolean active = null;
   private List stateChanges = null;
   private List reports = null;
   private HashMap participants = null;
   private HashMap participantPseudonyms = null;

   public Event()
   {
      this.eid = 0;
      this.eventtime = 0;
      this.modifytime = 0;
      this.headline = "";
      this.visible = new Boolean(false);
      this.active = new Boolean(false);
      this.stateChanges = new Vector();
      this.reports = new Vector();
      this.participantPseudonyms = new HashMap();
      this.participants = new HashMap();
      clean();
      newEventCache.add(this);
   }
   
   /** Create and event and set its event ID */
   Event(int eid)
   {
      this.eid = eid;
   }

   /** Create and event and set its event ID */
   Event(Integer eid)
   {
      this.eid = eid.intValue();
   }

   /** returns the database eid */
   int getEid() { return eid; }

   /**
    * Retrieve a mapping of participant indices to assassins for this event.
    * Each participant is assigned a unique index for the event so that it
    * can be used in [P1] tags and so on without having to worry about what
    * happens if other assassins are associated with or dissociated from
    * the event.
    * @return A Map of participant indices to Assassins.
    */
   Map getParticipants() throws AssassinSQLException
   {
      if (null == participants) readDetails(READPARTICIPANTS);
      return (Map) participants.clone();
   }
 
   /**
    * Retrieve a mapping of participant indices to pseudonyms for this event.
    * Each participant is assigned a unique index for the event so that it
    * can be used in [P1] tags and so on without having to worry about what
    * happens if other assassins are associated with or dissociated from
    * the event.
    * @return A Map of participant indices to Pseudonyms.
    */
   Map getParticipantPseudonyms() throws AssassinSQLException
   {
      if (null == participants) readDetails(READPARTICIPANTS);
      return (Map) participantPseudonyms.clone();
   }
 
   /**
    * Retrieve a mapping of participant indices to assassins for this event.
    * Each participant is assigned a unique index for the event so that it
    * can be used in [P1] tags and so on without having to worry about what
    * happens if other assassins are associated with or dissociated from
    * the event.
    * @return A Map of participant indices to AssassinViewers.
    */
   public Map getParticipantViewers() throws AssassinSQLException
   {
      if (null == participants) readDetails(READPARTICIPANTS);
      Map m = new HashMap();
      Iterator i = participants.keySet().iterator();
      while (i.hasNext()) {
         Integer j = (Integer) i.next();
         Assassin a = (Assassin) participants.get(j);
         m.put(j, a.getViewer());
      }
      return m;
   }
   
   /*
    * Mark a new assassin as being involved in this event.
    * @param whom The assassin to add to the event.
    * @param pseud An object representing the pseudonym to be used for this
    *    assassin in the event headline.
    * @return The participant index assigned to the newly-added participant.
    * /
   public int addParticipant(Assassin whom, Pseudonym pseud) throws AssassinSQLException
   {
      if (null == participants) readDetails(READPARTICIPANTS);
      setDirty();
      int i = participants.size();
      participants.put(new Integer(i+1), whom);
      participantPseudonyms.put(new Integer(i+1), pseud);
      return i+1;
   }*/

   /**
    * Determine which assassin a given participant index refers to.
    * @param which The participant index to ask about.
    * @return The object representing the assassin in question.
    */
   Assassin getParticipant(int which) throws AssassinSQLException
   {
      if (null == participants) readDetails(READPARTICIPANTS);
      return (Assassin) participants.get(new Integer(which));
   }
   
   /**
    * Determine which assassin a given participant index refers to.
    * @param which The participant index to ask about.
    * @return A viewer for the assassin in question.
    */
   public AssassinViewer getParticipantViewer(int which) throws AssassinSQLException
   {
      if (null == participants) readDetails(READPARTICIPANTS);
      return (AssassinViewer) ((Assassin) participants.get(new Integer(which))).getViewer();
   }
   
   /**
    * Retrieve which pseudonym is being used by a particular participant.
    * @param which The participant index to ask about.
    * @return An object representing the pseudonym being used.
    */
   Pseudonym getParticipantPseudonym(int which) throws AssassinSQLException
   {
      if (null == participants) readDetails(READPARTICIPANTS);
      return (Pseudonym) participantPseudonyms.get(new Integer(which));      
   }
   /**
    * Retrieve which pseudonym is being used by a particular participant.
    * @param which The participant index to ask about.
    * @return An object representing the pseudonym being used.
    */
   public String getParticipantPseudonymViewer(int which) throws AssassinSQLException
   {
      if (null == participants) readDetails(READPARTICIPANTS);
      return ((Pseudonym) participantPseudonyms.get(new Integer(which))).getValue();      
   }
   
   /**
    * Set an assassin/pseudonym as a participant
    * @param which The participant index to set/change.
    * @param whom The assassin to set.
    * @param pseud Which of the new assassin's pseudonyms to use for this event.
    */
   public void setParticipant(int which, Assassin whom, Pseudonym pseud)
   {
      setDirty();
      participants.put(new Integer(which), whom);
      participantPseudonyms.put(new Integer(which), pseud);
   }

   /**
    * Change the pseudonym to be used by a particular participant.
    * @param which The index of the participant whose pseudonym is being
    *    changed.
    * @param pseud The new pseudonym to use.
    */
   public void setParticipantPseudonym(int which, Pseudonym pseud) throws AssassinSQLException
   {
      if (null == participants) readDetails(READPARTICIPANTS);
      setDirty();
      participantPseudonyms.put(new Integer(which), pseud);
   }

   /**
    * Removes a participant from being associated with this event. Note that
    * the old participant index will not be reused if new participants are
    * added in future.
    * @param a The participant to remove.
    */
   public void removeParticipant(Assassin a) throws AssassinSQLException
   {
      if (null == participants) readDetails(READPARTICIPANTS);
      setDirty();
      Iterator i = participants.keySet().iterator();
      while (i.hasNext()) {
         Object k = i.next();
         if (participants.get(k).equals(a)) {
            participants.remove(k);
            participantPseudonyms.remove(k);
         }
      }
   }

   /**
    * Removes a participant from being associated with this event. Note that
    * the old participant index will not be reused if new participants are
    * added in future.
    * @param a The participant to remove.
    */
   public void removeParticipant(AssassinViewer a) throws AssassinSQLException
   {
      removeParticipant((Assassin) a.getEntity());
   }

   /**
    * Removes a participant from being associated with this event. Note that
    * the old participant index will not be reused if new participants are
    * added in future.
    * @param which The index of the participant to remove.
    */
   public void removeParticipant(int which) throws AssassinSQLException
   {
      if (null == participants) readDetails(READPARTICIPANTS);
      setDirty();
      participants.remove(new Integer(which));
      participantPseudonyms.remove(new Integer(which));
   }

   /**
    */
   public int getMaxParticipantID() throws AssassinSQLException
   {
      if (null == participants) readDetails(READPARTICIPANTS);
      Iterator i = participants.keySet().iterator();
      int max = 0;
      while (i.hasNext()) {
         int j = ((Integer) i.next()).intValue();
         if ( j > max ) max = j;
      }
      return max;
   }


   /**
    * Remove a report from this event.
    *
    * @param r The report to remove
    */
   public void removeReport(Report r) throws AssassinSQLException
   {
      if (null == reports) readDetails(READREPORTS);
      setDirty();
      reports.remove(r);
      r.setEvent(null);
   }
   /**
    * Add a report to this event.
    *
    * @param r The report to add
    */
   public void addReport(Report r) throws AssassinSQLException
   {
      if (null == reports) readDetails(READREPORTS);
      setDirty();
      reports.add(r);
      r.setEvent(this);
   }

   /**
    * Get a list of all the reports attached to this event. To change the
    * event's reports, use {@link addReport(Report r)}.
    * @return A list of reports attached to the event.
    */
   List getReports()  throws AssassinSQLException
   { 
      if (null == reports) readDetails(READREPORTS);
      return reports; 
   }

   /**
    * Get a list of state changes attached to this event.
    * @return A List of StateChanges attached to the event.
    */
   public List getStateChanges() throws AssassinSQLException
   {
      if (null == stateChanges) readDetails(READSTATECHANGES);
      return stateChanges;
   }

   /**
    * Attach a new state change to this event.
    * @param sc The state change to attach.
    */
   public void addStateChange(StateChange sc) throws AssassinSQLException
   {
      if (null == stateChanges) readDetails(READSTATECHANGES);
      setDirty();
      stateChanges.add(sc);
      sc.setEvent(this);
   }

   /**
    * Remove a state change from this event.
    * @param sc The state change to remove.
    */
   public void removeStateChange(StateChange sc) throws AssassinSQLException
   {
      if (null == stateChanges) readDetails(READSTATECHANGES);
      setDirty();
      stateChanges.remove(sc);
      sc.setEvent(null);
   }

   /**
    * Retrieve the time at which this event occurs.
    * @return The time at which the event occurs (in seconds since the epoch).
    */
   public long getEventTime()   throws AssassinSQLException
   {
      if (0 == eventtime) readDetails(READBUILTIN);
      return eventtime;
   }

   /**
    * Retrieve the time at which this event was last modified.
    * @return The time at which the event was last modified (in seconds since the epoch).
    */
   public long getModifiedTime()   throws AssassinSQLException
   {
      if (0 == modifytime) readDetails(READBUILTIN);
      return modifytime;
   }
   
   /**
    * Set a new time for this event.
    * @param time The new time for the event (in seconds since the epoch).
    */
   public void setEventTime(long time)
   {
      setDirty();
      this.eventtime = time;
   }

   /**
    * Set a new time for this event.
    * @param time The new time for the event (in seconds since the epoch).
    */
   public void setModifiedTime(long time)
   {
      setDirty();
      this.modifytime = time;
   }

   /**
    * Retrieve the headline for this event.
    * @return The headline string for the event.
    */
   public String getHeadline() throws AssassinSQLException
   {
      if (null == headline) readDetails(READBUILTIN);
      return headline;
   }
   /**
    * Set a new headline for this event.
    * @param headline The new headline.
    */
   public void setHeadline(String headline)
   {
      setDirty();
      this.headline = headline;
   }

   /**
    * Get the visibility of the event.
    * This defines whether it is listed in news pages
    *
    * @return Whether the event should be visible
    */
   public boolean isVisible() throws AssassinSQLException
   {
      if (null == visible) readDetails(READBUILTIN);
      return visible.booleanValue();
   }

   /**
    * Sets the visibility of the event
    *
    * @param visible The visibility state of the Event
    */
   public void setVisible(boolean visible)
   {
      setDirty();
      this.visible = new Boolean(visible);
   }

   /**
    * Gets the active state of the event.
    * If this is false, the event has no effect
    *
    * @return True if the event should have an effect
    */
   public boolean isActive() throws AssassinSQLException
   {
      if (null == active) readDetails(READBUILTIN);
      return active.booleanValue();
   }
   
   /**
    * Sets the active state of the event
    *
    * @param active The desired active state of the event
    */
   public void setActive(boolean active)
   {
      setDirty();
      this.active = new Boolean(active);
   }

   /**
    * Called by the database backend to serialise the object
    */
   public String toString()
   {
      return "__::org.ucam.srcf.assassins.Event("+eid+")";
   }

   /**
    * Reads in the details of the object from the database
    *
    * @throws AssassinSQLException If there is a problem with the database access.
    */
   void readDetails() throws AssassinSQLException
   {
      readDetails(READALL);
   }
   /**
    * Reads in the details of the object from the database
    *
    * @param mode 0 = read all, 1 = read builtins, 2 = read statechanges, 3 = read reports, 4 = read participants
    * @throws AssassinSQLException If there is a problem with the database access.
    */
   void readDetails(int mode) throws AssassinSQLException
   {
      if (dirty()) throw new ObjectModifiedException();
      
      if (0 == eid) throw new NoDatabaseAssociativityException();

      // read event details
      try {
         String query;
         ResultSet r;
         if (READALL == mode || READBUILTIN == mode) {
            if (Debug.debug) Debug.print(Event.class, "Reading Event Details");
            query = "SELECT headline, time, modified, active, visible FROM " + DatabaseBackend.getPrefix() +"_events WHERE eid = '"+eid+"'";
            r = DatabaseBackend.query(query);
            if (!r.next()) throw new NoDatabaseAssociativityException();
            this.headline = r.getString("headline");
            this.eventtime = r.getLong("time");
            this.modifytime = r.getLong("modified");
            this.active = new Boolean(r.getBoolean("active"));
            this.visible = new Boolean(r.getBoolean("visible"));
            r.close();
         }

         // read associated reports 
         if (READALL == mode || READREPORTS == mode) {
            if (Debug.debug) Debug.print(Event.class, "Reading Reports");
            query = "SELECT rid FROM " + DatabaseBackend.getPrefix() + "_reports WHERE event = '"+eid+"'";
            r = DatabaseBackend.query(query);
            reports = new Vector();
            while (r.next()) {
               reports.add(Report.getByRid(r.getInt("rid")));
            }
            r.close();
         }

         if (READALL == mode || READSTATECHANGES == mode) {
            // read associated state changes
            if (Debug.debug) Debug.print(Event.class, "Reading State Changes");
            query = "SELECT scid FROM " + DatabaseBackend.getPrefix() + "_statechanges WHERE event = '"+eid+"'";
            r = DatabaseBackend.query(query);
            stateChanges = new Vector();
            while (r.next()) {
               stateChanges.add(StateChange.getByScid(r.getInt("scid")));
            }
            r.close();
         }

         // read associated assassins/pseudonyms
         if (READALL == mode || READPARTICIPANTS == mode) {
            if (Debug.debug) Debug.print(Event.class, "Reading Participants");
            query = "SELECT aid, psid, pindex FROM " + DatabaseBackend.getPrefix() + "_eventparticipants WHERE eid = '"+eid+"'";
            r = DatabaseBackend.query(query);
            participants = new HashMap();
            participantPseudonyms = new HashMap();
            while (r.next()) {
               Assassin a = Assassin.getByAid(r.getInt("aid"));
               Pseudonym p = Pseudonym.getByPsid(r.getInt("psid"));
               Integer pindex = new Integer (r.getInt("pindex"));
               participants.put(pindex, a);
               participantPseudonyms.put(pindex, p);
            }
            r.close();
         }
      } catch (SQLException SQLe) { throw new AssassinSQLException(SQLe); }
   }

   /**
    * Writes the details of the object to the database
    *
    * @throws AssassinSQLException If there is a problem with the database access.
    */
   void writeDetails() throws AssassinSQLException
   {
      if (!dirty()) return;
      clean();

      try { 
         // create new
         if (0 == eid) {

            // insert the event
            String query = "INSERT INTO " + DatabaseBackend.getPrefix() + "_events (headline, time, modified, active, visible) VALUES ('"+Tools.escape(headline)+"','"+eventtime+"','"+modifytime+"','"+(active.booleanValue()?"1":"0")+"','"+(visible.booleanValue()?"1":"0")+"')";
            ResultSet r = DatabaseBackend.query(query);            
            r.close();
            query = "SELECT  LAST_INSERT_ID() FROM " + DatabaseBackend.getPrefix() + "_events";
            // get auto-assigned pid
            r = DatabaseBackend.query(query);
            if (Debug.debug) Debug.print(autoumpire.class, "Getting ID");
            if (!r.next()) throw new AssassinSQLException("Oh Ek! That shouldn't happen?!");
            eid = r.getInt(1);

            if (Debug.debug) Debug.print(autoumpire.class, "Saving Reports");
            // write out reports
            if (null != reports) {
               Iterator i = reports.iterator();
               while (i.hasNext())
                  ((Report) i.next()).writeDetails();
            }

            if (Debug.debug) Debug.print(autoumpire.class, "Saving State Changes");
            // write out state changes
            if (null != stateChanges) {
               Iterator i = stateChanges.iterator();
               while (i.hasNext())
                  ((StateChange) i.next()).writeDetails();
            }

            if (Debug.debug) Debug.print(autoumpire.class, "Saving Participants");
            // write out participant/pseudonym sets
            if (null != participants || null != participantPseudonyms) {
               Iterator i = participants.keySet().iterator();
               while (i.hasNext()) {
                  Integer J = (Integer) i.next();
                  int j = J.intValue();
                  if (Debug.debug) Debug.print(Event.class, "Saving participant " + J);
		  query = "INSERT INTO " + DatabaseBackend.getPrefix() + "_eventparticipants (aid, eid, psid, pindex) VALUES ('";
                  if (null != participants.get(J))
                     query += ((Assassin)participants.get(J)).getAid();
                  else
                     query += 0;
                  query += "','"+eid+"','";
                  if (null != participantPseudonyms.get(J))
                     query += ((Pseudonym)participantPseudonyms.get(J)).getPsid();
                  else
                     query += 0;
                  query += "','"+j+"')";
                  DatabaseBackend.query(query);
               }
            }
         }

         // update existing
         else {

            // update ours
            String query = "UPDATE " + DatabaseBackend.getPrefix() + "_events SET ";
            if (null != this.headline) query += "headline = '" + Tools.escape(this.headline) + "', ";
            if (0 != this.eventtime) query += "time = '" + this.eventtime + "', ";
            if (0 != this.modifytime) query += "modified = '" + this.modifytime + "', ";
            if (null != this.active) query += "active = '" + (this.active.booleanValue()?"1":"0") + "', ";
            if (null != this.visible) query += "visible = '" + (this.visible.booleanValue()?"1":"0") + "', ";
            query += " eid = '"+eid+"' WHERE eid = '" + eid +"'";
            DatabaseBackend.query(query);

            // write out reports
            if (null != reports) {
               Iterator i = reports.iterator();
               while (i.hasNext()) {
                  Report rep = ((Report) i.next());
                  if (Debug.debug) Debug.print(this, "Writing report: "+rep);
                  rep.writeDetails();
               }
            }

            // write out state changes
            if (null != stateChanges) {
               Iterator i = stateChanges.iterator();
               while (i.hasNext())
                  ((StateChange) i.next()).writeDetails();
            }

            if (null != participants) {
               // write out participant/pseudonym sets
               query = "DELETE FROM " + DatabaseBackend.getPrefix() + "_eventparticipants WHERE eid = '"+eid+"'";
               DatabaseBackend.query(query);
               // write out participant/pseudonym sets
               if (null != participants || null != participantPseudonyms) {
                  Iterator i = participants.keySet().iterator();
                  while (i.hasNext()) {
                     Integer J = (Integer) i.next();
                     int j = J.intValue();
                     if (Debug.debug) Debug.print(Event.class, "Saving participant " + J);
                     query = "INSERT INTO " + DatabaseBackend.getPrefix() + "_eventparticipants (aid, eid, psid, pindex) VALUES ('";
                     if (null != participants.get(J))
                        query += ((Assassin)participants.get(J)).getAid();
                     else
                        query += 0;
                     query += "','"+eid+"','";
                     if (null != participantPseudonyms.get(J))
                        query += ((Pseudonym)participantPseudonyms.get(J)).getPsid();
                     else
                        query += 0;
                     query += "','"+j+"')";
                     DatabaseBackend.query(query);
                  }
               }
            }
         }
      } catch (SQLException SQLe) { throw new AssassinSQLException(SQLe); }

   }

   protected void setDirty()
   {
      try {
         this.readDetails(READALL);
      } catch(AssassinSQLException ASQLe) {}
      super.setDirty();
      modifytime = System.currentTimeMillis();
   }

   public Viewer getViewer()
   {
      return new EventViewer(this);
   }
   
   public String getPrintable() throws AssassinSQLException
   {
      if (null == participantPseudonyms) readDetails(READPARTICIPANTS);
      Map m = new HashMap();
      Iterator i = participantPseudonyms.keySet().iterator();
      while (i.hasNext()) {
         Integer n = (Integer) i.next();
         Pseudonym p = (Pseudonym) participantPseudonyms.get(n);
         if (null != p) {
            String s = p.getValue();
            if (s.length() > 20) s = s.substring(0,20)+"...";
            m.put("P"+n, s);
            m.put("p"+n, s);
         }
         m.put("N"+n, ((Assassin) participants.get(n)).getPlayer().getName());
         m.put("n"+n, ((Assassin) participants.get(n)).getPlayer().getName());
      }
      return "[" + Tools.formatTime(getEventTime()) +"] " + Tools.replace(getHeadline(), m);
   }

}



