/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.exceptions.NoDatabaseAssociativityException;
import org.ucam.srcf.assassins.exceptions.NonExistentEntityException;
import org.ucam.srcf.assassins.exceptions.NotSerialisedEntityException;
import org.ucam.srcf.assassins.exceptions.ObjectModifiedException;
/**
 * Base class of all changes to the gamestate that vanilla knows about.
 * Note that all state change classes should be <b>immutable</b> once
 * constructed or else <b>really terrible things may happen</b>.
 * 
 * @author Matthew Johnson <autoumpire@matthew.ath.cx>
 */

public abstract class StateChange extends CanonicalEntity
{
   /** Cache of StateChange objects */
   private static Map stateChangeCache = new HashMap();
   /** Cache of new StateChange objects */
   private static List newStateChangeCache = new Vector();
   /** Cache of Player objects */
   private static Map playerCache = new HashMap();
   private static boolean preCacheState = false;


   private static List loadChangesFromResultSet(ResultSet r) throws SQLException
   {
      List l = new Vector();
      while (r.next()) {

         int scid = r.getInt("scid");
         Integer SCID = new Integer(scid);

         // check the cache and add as necessary
         if (stateChangeCache.containsKey(SCID)) {
            l.add(stateChangeCache.get(SCID));
         } else {
            String sccl = r.getString("class");
            if (Debug.debug) Debug.print(StateChange.class, "Reflecting on "+ sccl);
            StateChange sc = null;
            try {
               Class c = Class.forName(sccl);
               Constructor con = c.getDeclaredConstructor(new Class[] {Integer.class});
               sc = (StateChange) con.newInstance(new Object[] {SCID});
            } 
            catch (NoSuchMethodException NSMe) { throw new AssassinSQLException(); }
            catch (InstantiationException Ie) { throw new AssassinSQLException(); }
            catch (IllegalAccessException IAe) { throw new AssassinSQLException(); }
            catch (ClassNotFoundException CNFe) { throw new AssassinSQLException(); }
            catch (InvocationTargetException ITe) { throw new AssassinSQLException(); }
            String temps;
            sc.event = Event.getByID(r.getInt("event"));
            sc.pluginName = r.getString("pluginName");
            sc.field = r.getString("field");
            sc.ass = Assassin.getByAid(r.getInt("assassin"));
            temps = r.getString("value");

            // check if its a canonical entity, otherwise its a string.
            try {
               sc.value = CanonicalEntity.deSerialise(temps);
            } catch (NotSerialisedEntityException NSEe) {
               sc.value = temps;
               if (Debug.debug) Debug.print(StateChange.class, "Found a string: "+temps);
               if (Debug.debug) Debug.print(StateChange.class, "Cause: "+NSEe.getMessage());
            }
            stateChangeCache.put(SCID, sc);
            l.add(sc);
         }
      }
      return l;
   }

   /**
    * Loads all the objects into memory frm the database in one query
    */
   static void preCache() throws AssassinSQLException
   {
      if (preCacheState) return;
      preCacheState = true;
      ResultSet r = DatabaseBackend.query("SELECT scid, class, event, assassin, pluginName, field, value FROM "+DatabaseBackend.getPrefix()+"_statechanges");
      try {
         loadChangesFromResultSet(r);
         r.close();
      } catch (SQLException sqle) {
         throw new AssassinSQLException();
      }
   }
   
   private static List getByQuery(String query) throws AssassinSQLException
   {
      List stateChanges;
      try {
         ResultSet r = DatabaseBackend.query(query);
         stateChanges = loadChangesFromResultSet(r);
         r.close();
      } catch (SQLException sqle) {
         sqle.printStackTrace();
         throw new AssassinSQLException();
      }
      return stateChanges;
   }

   /** 
    * Returns all the active state changes on a particular assassin up to given time
    *
    * @param a The Assassin you wish to search for
    * @param t The time to search up until
    * @return A List of StateChange Objects
    * @throws AssassinSQLException
    */
   public static List getActiveByAssassin(Assassin a, long t) throws AssassinSQLException
   {
      String sct = DatabaseBackend.getPrefix() + "_statechanges sc";
      String evt = DatabaseBackend.getPrefix() + "_events ev";
      String query = "SELECT scid, class, event, assassin, pluginName, field, value FROM " +sct+" INNER JOIN "+evt+ " ON ev.eid = sc.event WHERE sc.assassin = '" + a.getAid() + "' AND ev.active = '1' AND ev.time <= '"+t+"' ORDER BY ev.time";
      return getByQuery(query);
   }


   /** 
    * Returns all the state changes on a particular assassin 
    *
    * @param a The Assassin you wish to search for
    * @return A List of StateChange Objects
    * @throws AssassinSQLException
    */
   public static List getByAssassin(Assassin a) throws AssassinSQLException
   {
      String sct = DatabaseBackend.getPrefix() + "_statechanges sc";
      String evt = DatabaseBackend.getPrefix() + "_events ev";
      String query = "SELECT scid, class, event, assassin, pluginName, field, value FROM " +sct+","+evt+ " WHERE sc.assassin = '" + a.getAid() + "' AND ev.eid = sc.event ORDER BY ev.time";
      return getByQuery(query);
   }

   /** 
    * Returns all the state changes on a particular assassin .
    * (takes an AssassinViewer)
    *
    * @param a The Assassin you wish to search for
    * @return A List of StateChange Objects
    * @throws AssassinSQLException
    */
   public static List getByAssassin(AssassinViewer a) throws AssassinSQLException
   {
      String sct = DatabaseBackend.getPrefix() + "_statechanges sc";
      String evt = DatabaseBackend.getPrefix() + "_events ev";
      String query = "SELECT scid, class, event, assassin, pluginName, field, value FROM " +sct+","+evt+ " WHERE assassin = '" + ((Assassin)a.getEntity()).getAid() + "' AND ev.eid = sc.event ORDER BY ev.time";
      return getByQuery(query);
   }

   /** 
    * Returns the value for a particular field on a particular assassin at the given time.
    * (takes an Assassin)
    *
    * @param a The Assassin you wish to search for
    * @param pluginName Name of the plugin
    * @param field Field to select on
    * @param t time to get the value for
    * @return A List of either 1 or 0 StateChanges if this field has been set
    * @throws AssassinSQLException
    */
   public static List getFieldValueAtTime(Assassin a, String pluginName, String field, long t) throws AssassinSQLException
   {
      String sct = DatabaseBackend.getPrefix() + "_statechanges sc";
      String evt = DatabaseBackend.getPrefix() + "_events ev";
      String query = "SELECT scid, class, event, assassin, pluginName, field, value FROM " +sct+" INNER JOIN"+evt+ " ON ev.eid = sc.event WHERE sc.assassin = '" + a.getAid() + "' AND sc.pluginName = '"+pluginName+"' AND sc.field = '"+field+"' and time <= '"+t+"' ORDER BY ev.time DESC LIMIT 1";
      return getByQuery(query);
   }

   /** 
    * Returns the value for a particular field on a particular assassin at the given time.
    * (takes an AssassinViewer)
    *
    * @param a The Assassin you wish to search for
    * @param pluginName Name of the plugin
    * @param field Field to select on
    * @param t time to get the value for
    * @return A List of either 1 or 0 StateChanges if this field has been set
    * @throws AssassinSQLException
    */
   public static List getFieldValueAtTime(AssassinViewer a, String pluginName, String field, long t) throws AssassinSQLException
   {
      String sct = DatabaseBackend.getPrefix() + "_statechanges sc";
      String evt = DatabaseBackend.getPrefix() + "_events ev";
      String query = "SELECT scid, class, event, assassin, pluginName, field, value FROM " +sct+" INNER JOIN"+evt+ " ON ev.eid = sc.event WHERE sc.assassin = '" + ((Assassin) a.getEntity()).getAid() + "' AND sc.pluginName = '"+pluginName+"' AND sc.field = '"+field+"' and time <= '"+t+"' ORDER BY ev.time DESC LIMIT 1";
      return getByQuery(query);
   }
   
   /** 
    * Returns the value for a particular field on a every assassin at the given time.
    *
    * @param pluginName Name of the plugin
    * @param field Field to select on
    * @param t time to get the value for
    * @return A List of either 1 or 0 StateChanges if this field has been set
    * @throws AssassinSQLException
    */
   public static List getEveryonesFieldValueAtTime(String pluginName, String field, long t) throws AssassinSQLException
   {
      /*
       * MySQL is a bit anal here about using table aliases when locking tables.
       * Yes you do need the names explicitly.
       * - mjj29
       */
      String dbp = DatabaseBackend.getPrefix();
      String query = "create temporary table "+dbp+"_tmp (time bigint, assassin int, pluginName varchar(20), field varchar(20))";
      DatabaseBackend.query(query);
      
      query = "lock tables "+dbp+"_events read, "+dbp+"_statechanges read;";
      DatabaseBackend.query(query);
      
      query = "insert into "+dbp+"_tmp select max(time), assassin, pluginName, field from "+dbp+"_statechanges inner join "+dbp+"_events ON "+dbp+"_statechanges.event = "+dbp+"_events.eid where "+dbp+"_events.time <= '"+t+"' and "+dbp+"_statechanges.pluginName='"+pluginName+"' and "+dbp+"_statechanges.field='"+field+"' group by assassin";
      DatabaseBackend.query(query);
      
      query = "select "+dbp+"_statechanges.scid, "+dbp+"_statechanges.class, "+dbp+"_statechanges.pluginName, "+dbp+"_statechanges.field, "+dbp+"_statechanges.value from "+dbp+"_events, "+dbp+"_statechanges, "+dbp+"_tmp where "+dbp+"_events.time <= '"+t+"' and "+dbp+"_events.eid="+dbp+"_statechanges.event and "+dbp+"_tmp.assassin="+dbp+"_statechanges.assassin and "+dbp+"_tmp.pluginName="+dbp+"_statechanges.pluginName and "+dbp+"_tmp.field="+dbp+"_statechanges.field and "+dbp+"_events.time="+dbp+"_tmp.time and "+dbp+"_statechanges.pluginName='"+pluginName+"' and "+dbp+"_statechanges.field='"+field+"'";
      ResultSet r = DatabaseBackend.query(query);
      
      query = "unlock tables";
      DatabaseBackend.query(query);
      
      query = "drop table "+dbp+"_tmp";
      DatabaseBackend.query(query);
      
      List stateChanges;
      try {
         stateChanges = loadChangesFromResultSet(r);
         r.close();
      } catch (SQLException sqle) {
         throw new AssassinSQLException();
      }
      return stateChanges;
   }
   
   /** 
    * Returns the latest value for every field on an assassin at a given time
    *
    * @param a The Assassin for the StateChange type to filter on
    * @param t The time to get the values
    * @return A List of StateChange Objects
    * @throws AssassinSQLException
    */
   public static List getFieldValuesAtTime(Assassin a, long t) throws AssassinSQLException
   {
      /*
       * MySQL is a bit anal here about using table aliases when locking tables.
       * Yes you do need the names explicitly.
       * - mjj29
       */
      String dbp = DatabaseBackend.getPrefix();
      String query = "create temporary table "+dbp+"_tmp (time bigint, assassin int, pluginName varchar(20), field varchar(20))";
      DatabaseBackend.query(query);
      
      query = "lock tables "+dbp+"_events read, "+dbp+"_statechanges read;";
      DatabaseBackend.query(query);
      
      query = "insert into "+dbp+"_tmp select max(time), assassin, pluginName, field from "+dbp+"_statechanges inner join "+dbp+"_events on "+dbp+"_statechanges.event = "+dbp+"_events.eid where "+dbp+"_statechanges.assassin='"+a.getAid()+"' and "+dbp+"_events.time <= '"+t+"' group by pluginName, field";
      DatabaseBackend.query(query);
      
      query = "select "+dbp+"_statechanges.scid, "+dbp+"_statechanges.class, "+dbp+"_statechanges.pluginName, "+dbp+"_statechanges.field, "+dbp+"_statechanges.value from "+dbp+"_events, "+dbp+"_statechanges, "+dbp+"_tmp where "+dbp+"_statechanges.assassin='"+a.getAid()+"' and "+dbp+"_events.time <= '"+t+"' and "+dbp+"_events.eid="+dbp+"_statechanges.event and "+dbp+"_tmp.assassin="+dbp+"_statechanges.assassin and "+dbp+"_tmp.pluginName="+dbp+"_statechanges.pluginName and "+dbp+"_tmp.field="+dbp+"_statechanges.field and "+dbp+"_events.time="+dbp+"_tmp.time";
      ResultSet r = DatabaseBackend.query(query);
      
      query = "unlock tables";
      DatabaseBackend.query(query);
      
      query = "drop table "+dbp+"_tmp";
      DatabaseBackend.query(query);
      
      List stateChanges;
      try {
         stateChanges = loadChangesFromResultSet(r);
         r.close();
      } catch (SQLException sqle) {
         throw new AssassinSQLException();
      }
      return stateChanges;

   }

   /** 
    * Returns the latest value for every field on an assassin at a given time
    * (takes an AssassinViewer)
    *
    * @param a The Assassin for the StateChange type to filter on
    * @param t The time to get the values
    * @return A List of StateChange Objects
    * @throws AssassinSQLException
    */
   public static List getFieldValuesAtTime(AssassinViewer a, long t) throws AssassinSQLException
   {
      return getFieldValuesAtTime((Assassin) a.getEntity(), t);
   }

   /** 
    * Returns the latest value for every field on every assassin at a given time
    *
    * @return A List of StateChange Objects
    * @throws AssassinSQLException
    */
   public static List getFieldValuesAtTime(long t) throws AssassinSQLException
   {
      /*
       * MySQL is a bit anal here about using table aliases when locking tables.
       * Yes you do need the names explicitly.
       * - mjj29
       */
      String dbp = DatabaseBackend.getPrefix();
      String query = "create temporary table "+dbp+"_tmp (time bigint, assassin int, pluginName varchar(20), field varchar(20))";
      DatabaseBackend.query(query);
      
      query = "lock tables "+dbp+"_events read, "+dbp+"_statechanges read;";
      DatabaseBackend.query(query);
      
      query = "insert into "+dbp+"_tmp select max(time), assassin, pluginName, field from "+dbp+"_statechanges inner join "+dbp+"_events ON "+dbp+"_statechanges.event = "+dbp+"_events.eid where "+dbp+"_events.time <= '"+t+"' group by assassin, pluginName, field";
      DatabaseBackend.query(query);
      
      query = "select "+dbp+"_statechanges.scid, "+dbp+"_statechanges.class, "+dbp+"_statechanges.pluginName, "+dbp+"_statechanges.field, "+dbp+"_statechanges.value from "+dbp+"_events, "+dbp+"_statechanges, "+dbp+"_tmp where "+dbp+"_events.time <= '"+t+"' and "+dbp+"_events.eid="+dbp+"_statechanges.event and "+dbp+"_tmp.assassin="+dbp+"_statechanges.assassin and "+dbp+"_tmp.pluginName="+dbp+"_statechanges.pluginName and "+dbp+"_tmp.field="+dbp+"_statechanges.field and "+dbp+"_events.time="+dbp+"_tmp.time";
      ResultSet r = DatabaseBackend.query(query);
      
      query = "unlock tables";
      DatabaseBackend.query(query);
      
      query = "drop table "+dbp+"_tmp";
      DatabaseBackend.query(query);
      
      List stateChanges;
      try {
         stateChanges = loadChangesFromResultSet(r);
         r.close();
      } catch (SQLException sqle) {
         throw new AssassinSQLException();
      }
      return stateChanges;
   }

   /** 
    * Returns all the death/ressurrection state changes by time t
    *
    * @param t The time for the StateChange type to filter on
    * @return A List of StateChange Objects
    * @throws AssassinSQLException
    */
   public static List getVitalStatusUntilTime(long t) throws AssassinSQLException
   {
      String sct = DatabaseBackend.getPrefix() + "_statechanges sc";
      String evt = DatabaseBackend.getPrefix() + "_events ev";
      String query = "SELECT scid, class, event, assassin, pluginName, field, value FROM " +sct+" INNER JOIN "+evt+ " ON ev.eid = sc.event WHERE (class = 'org.ucam.srcf.assassins.Resurrection' OR class = 'org.ucam.srcf.assassins.Death') AND ev.time <= '"+t+"' ORDER BY ev.time";
      return getByQuery(query);
   }

   /** 
    * Returns all the pseudonym-related state changes by time t
    *
    * @param t The time for the StateChange type to filter on
    * @return A List of StateChange Objects
    * @throws AssassinSQLException
    */
   public static List getPseudonymsAtTime(long t) throws AssassinSQLException
   {
      String sct = DatabaseBackend.getPrefix() + "_statechanges sc";
      String evt = DatabaseBackend.getPrefix() + "_events ev";
      String query = "SELECT scid, class, event, assassin, pluginName, field, value FROM " +sct+" INNER JOIN "+evt+ " ON ev.eid = sc.event WHERE (class = 'org.ucam.srcf.assassins.PseudonymGain' OR class = 'org.ucam.srcf.assassins.PseudonymLoss') AND ev.time <= '"+t+"' ORDER BY ev.time";
      return getByQuery(query);
   }


   /** 
    * Returns all the pseudonym-related state changes of a particular assassin by time t
    *
    * @param c The Class object for the StateChange type to filter on
    * @return A List of StateChange Objects
    * @throws AssassinSQLException
    */
   public static List getPseudonymByAssassin(Assassin a, long t) throws AssassinSQLException
   {
      String sct = DatabaseBackend.getPrefix() + "_statechanges sc";
      String evt = DatabaseBackend.getPrefix() + "_events ev";
      String query = "SELECT scid, class, event, assassin, pluginName, field, value FROM " +sct+" INNER JOIN "+evt+ " ON ev.eid = sc.event WHERE (class = 'org.ucam.srcf.assassins.PseudonymGain' OR class = 'org.ucam.srcf.assassins.PseudonymLoss') AND sc.assassin = '"+a.getAid()+"' AND ev.time <= '"+t+"' ORDER BY ev.time";
      return getByQuery(query);
   }

   /** 
    * Returns all the state changes of a particular class by the given time
    *
    * @param c The Class object for the StateChange type to filter on
    * @param time The time to filter before
    * @return A List of StateChange Objects
    * @throws AssassinSQLException
    */
   public static List getByClassAtTime(Class c, long time) throws AssassinSQLException
   {
      String sct = DatabaseBackend.getPrefix() + "_statechanges sc";
      String evt = DatabaseBackend.getPrefix() + "_events ev";
      String query = "SELECT scid, class, event, assassin, pluginName, field, value FROM " +sct+","+evt+ " WHERE class = '" + c.getName() + "' AND ev.eid = sc.event AND ev.time <= '"+time+"' ORDER BY ev.time";
      return getByQuery(query);
   }


   /** 
    * Returns all the state changes of a particular class
    *
    * @param c The Class object for the StateChange type to filter on
    * @return A List of StateChange Objects
    * @throws AssassinSQLException
    */
   public static List getByClass(Class c) throws AssassinSQLException
   {
      String sct = DatabaseBackend.getPrefix() + "_statechanges sc";
      String evt = DatabaseBackend.getPrefix() + "_events ev";
      String query = "SELECTcid, class, event, assassin, pluginName, field, value FROM " +sct+","+evt+ " WHERE class = '" + c.getName() + "' AND ev.eid = sc.event ORDER BY ev.time";
      return getByQuery(query);
   }

   /**
    * Gets all the active state changes between two times
    *
    * @param fromtime The time to search from
    * @param totime The time to search until
    * @return A List of StateChange Objects
    * @throws AssassinSQLException
    */
   public static List getActiveBetweenTimes(long fromtime, long totime) throws AssassinSQLException
   {
      String sct = DatabaseBackend.getPrefix() + "_statechanges sc";
      String evt = DatabaseBackend.getPrefix() + "_events ev";
      String query = "SELECT scid, class, event, assassin, pluginName, field, value FROM " +sct+" INNER JOIN "+evt+ " ON ev.eid = sc.event WHERE ev.time > '"+fromtime+"' AND ev.time <= '"+totime+"' AND ev.active='1' ORDER BY ev.time";
      return getByQuery(query);
   }

   /**
    * Gets all the active state changes before a time
    *
    * @param time The time to search for
    * @return A List of StateChange Objects
    * @throws AssassinSQLException
    */
   public static List getActiveBeforeTimeOrderedByTime(long time) throws AssassinSQLException
   {
      String sct = DatabaseBackend.getPrefix() + "_statechanges sc";
      String evt = DatabaseBackend.getPrefix() + "_events ev";
      String query = "SELECT scid, class, event, assassin, pluginName, field, value FROM " +sct+" INNER JOIN "+evt+ " ON ev.eid = sc.event WHERE ev.time <= '"+time+"' AND ev.active='1' ORDER BY ev.time";
      return getByQuery(query);
   }
    
   /**
    * Gets all the state changes for a particular field at time t
    *
    * @param s The field to search for
    * @param t The t to search at
    * @return A List of StateChange Objects
    * @throws AssassinSQLException
    */
   public static List getByFieldAtTime(String s, long t) throws AssassinSQLException
   {
      String sct = DatabaseBackend.getPrefix() + "_statechanges sc";
      String evt = DatabaseBackend.getPrefix() + "_events ev";
      String query = "SELECT scid, class, event, assassin, pluginName, field, value FROM " +sct+","+evt+ " WHERE field = '" + s + "' AND ev.eid = sc.event AND ev.time <= '"+t+"' ORDER BY ev.time";
      return getByQuery(query);
   }
   
   /**
    * Gets all the state changes 
    *
    * @return A List of StateChange Objects
    * @throws AssassinSQLException
    */
   public static List getAll() throws AssassinSQLException
   {
      String sct = DatabaseBackend.getPrefix() + "_statechanges sc";
      String evt = DatabaseBackend.getPrefix() + "_events ev";
      String query = "SELECT scid, class, event, assassin, pluginName, field, value FROM " +sct+","+evt+ " WHERE ev.eid = sc.event ORDER BY ev.time";
      return getByQuery(query);
   }
   
   /**
    * Gets all the state changes for a particular field
    *
    * @param s The field to search for
    * @return A List of StateChange Objects
    * @throws AssassinSQLException
    */
   public static List getByField(String s) throws AssassinSQLException
   {
      String sct = DatabaseBackend.getPrefix() + "_statechanges sc";
      String evt = DatabaseBackend.getPrefix() + "_events ev";
      String query = "SELECT scid, class, event, assassin, pluginName, field, value FROM " +sct+","+evt+ " WHERE field = '" + s + "' AND ev.eid = sc.event ORDER BY ev.time";
      return getByQuery(query);
   }
   
   /**
    * Gets all the (un)processed state changes before time t
    *
    * @param p Processed/Unprocessed
    * @param t Time
    * @return A List of StateChange Objects
    * @throws AssassinSQLException
    */
   public static List getProcessedAtTime(boolean p, long time) throws AssassinSQLException
   {
      String sct = DatabaseBackend.getPrefix() + "_statechanges sc";
      String evt = DatabaseBackend.getPrefix() + "_events ev";
      String query = "SELECT scid, class, event, assassin, pluginName, field, value FROM " +sct+","+evt+ " WHERE processed = '"+(p?"1":"0")+"' AND ev.eid = sc.event AND ev.time <= '"+time+"' ORDER BY ev.time";
      return getByQuery(query);
   }
      
   /**
    * Gets all the (un)processed state changes
    *
    * @param p Processed/Unprocessed
    * @return A List of StateChange Objects
    * @throws AssassinSQLException
    */
   public static List getProcessed(boolean p) throws AssassinSQLException
   {
      String sct = DatabaseBackend.getPrefix() + "_statechanges sc";
      String evt = DatabaseBackend.getPrefix() + "_events ev";
      String query = "SELECT scid, class, event, assassin, pluginName, field, value FROM " +sct+","+evt+ " WHERE processed = '"+(p?"1":"0")+"' AND ev.eid = sc.event ORDER BY ev.time";
      return getByQuery(query);
   }
      
   
   /** Returns a StateChange selected by Scid */
   static StateChange getByScid(int scid) throws NonExistentEntityException, AssassinSQLException
   {
      if (0 == scid) return null;
      StateChange sc;

      // check cache
      if (null == (sc = (StateChange) stateChangeCache.get(new Integer(scid)))) {
         // check database existence
         String query = "SELECT class FROM " + DatabaseBackend.getPrefix() + "_statechanges WHERE scid = '"+scid+"'";
         ResultSet r;
         try {
            r = DatabaseBackend.query(query);
            if (!r.next()) throw new NonExistentEntityException("Eek");
         } catch (SQLException e) {
            throw new AssassinSQLException();
         }
         // get the correct class of the state change
         try {
            if (Debug.debug) Debug.print(StateChange.class, "Reflecting on "+ r.getString("class"));
            Class c = Class.forName(r.getString("class"));
            Constructor con = c.getDeclaredConstructor(new Class[] {Integer.class});
            sc = (StateChange) con.newInstance(new Object[] {new Integer(scid)});
            r.close();
         }
         // re-throw all the exceptions, wrapped
         catch (ClassNotFoundException CNFe) { if (Debug.debug) Debug.print(StateChange.class, "Exception has occurred: "+CNFe); throw new NonExistentEntityException(CNFe); } 
         catch (ClassCastException CCe) { if (Debug.debug) Debug.print(StateChange.class, "Exception has occurred: "+CCe); throw new NonExistentEntityException(CCe); } 
         catch (NoSuchMethodException NSMe) { if (Debug.debug) Debug.print(StateChange.class, "Exception has occurred: "+NSMe); throw new NonExistentEntityException(NSMe); } 
         catch (SecurityException Se) { if (Debug.debug) Debug.print(StateChange.class, "Exception has occurred: "+Se); throw new NonExistentEntityException(Se); } 
         catch (InstantiationException Ie) { if (Debug.debug) Debug.print(StateChange.class, "Exception has occurred: "+Ie); throw new NonExistentEntityException(Ie); }
         catch (IllegalAccessException IAce) { if (Debug.debug) Debug.print(StateChange.class, "Exception has occurred: "+IAce); throw new NonExistentEntityException(IAce); }
         catch (IllegalArgumentException IAre) { if (Debug.debug) Debug.print(StateChange.class, "Exception has occurred: "+IAre); throw new NonExistentEntityException(IAre); }
         catch (InvocationTargetException ITe) { if (Debug.debug) Debug.print(StateChange.class, "Exception has occurred: "+ITe); throw new NonExistentEntityException(ITe); }
         catch (SQLException sqle) {if (Debug.debug) Debug.print(StateChange.class, "Exception has occurred: "+sqle); throw new AssassinSQLException(sqle);}

         // add to cache
         playerCache.put(new Integer(scid), sc);
      }
      return sc;

   }
  
   /** Called via reflection to get an instance of this by database ID */
   static Object getByID(Integer id) throws AssassinSQLException
   {
      return getByScid(id.intValue());
   }

   /**
    * Updates all modified StateChanges to the database
    */
   static void commitAll() throws AssassinSQLException
   {
      // iterate over all new StateChange objects
      while (newStateChangeCache.size() > 0) {
         StateChange a = (StateChange) newStateChangeCache.remove(0);
         a.writeDetails();
         stateChangeCache.put(new Integer(a.getScid()),a);
      }
      
      // iterate over all cached StateChange objects
      Iterator i = stateChangeCache.values().iterator();
      while (i.hasNext()) {
         StateChange sc = (StateChange) i.next();
         // if they are modified, then update
         if (sc.dirty()) sc.writeDetails();
      }

   }
   
   private int scid = 0;
   protected Assassin ass = null;
   protected String pluginName = null;
   protected String field = null;
   protected Object value = null;
   private Event event = null;
   
   private Boolean processed = null;
   
   void markProcessed() 
   {
      processed = new Boolean(true);
      setDirty();
   }
   
   public boolean isProcessed()  throws AssassinSQLException 
   {
      if (null == processed) readDetails();
      return processed.booleanValue();
   }
 
   /** Create a new StateChange */
   public StateChange()
   {
      processed = new Boolean(false);
      newStateChangeCache.add(this);
      clean();
   }

   StateChange(int scid)
   {
      this.scid = scid;
      clean();
   }

   int getScid() {
      return scid;
   }

   /**
    * Constructor to allocate the scid. 
    * This <b>MUST</b> be in all subclasses and <b>MUST</b> call this constructor.
    */
   StateChange(Integer scid)
   {
      this.scid = scid.intValue();
      clean();
   }
 
   /**
    * Retrieves the event that this state change is attached to.
    * @return The Event.
    */
   protected Event getEvent() throws AssassinSQLException
   {
      if (null == event && !dirty()) readDetails();
      return event;
   }
   
   /**
    * Retrieves a viewer for the event that this state change is attached to.
    * @return The EventViewer.
    */
   public EventViewer getEventViewer() throws AssassinSQLException
   {
      if (null == event && !dirty()) readDetails();
      return new EventViewer(event);
   }
   
   void setEvent(Event e)
   {
      this.event = e;
      setDirty();
   }

   Assassin getAssassin() throws AssassinSQLException
   {
      if (null == ass) readDetails();
      return ass;
   }
 
   public AssassinViewer getAssassinViewer() throws AssassinSQLException
   {
      if (null == ass) readDetails();
      return new AssassinViewer(ass);
   }
   
   public String getPluginName() throws AssassinSQLException
   {
      if (null == pluginName) readDetails();
      return pluginName;
   }
   
   public String getFieldName() throws AssassinSQLException
   {
      if (null == field) readDetails();
      return field;
   }
   
   public Object getValue() throws AssassinSQLException
   {
      if (null == value) readDetails();
      return value;
   }

   CanonicalEntity getEntity() throws AssassinSQLException
   {
      if (null == value) readDetails();
      if (value instanceof CanonicalEntity)
         return (CanonicalEntity) value;
      return null;
   }
 
   public CanonicalViewer getEntityViewer() throws AssassinSQLException
   {
      if (null == value) readDetails();
      if (value instanceof CanonicalEntity)
         return (CanonicalViewer) ((CanonicalEntity) value).getViewer();
      return null;
   }
 
   public String getStringValue() throws AssassinSQLException
   {
      if (null == value) readDetails();
      if (value instanceof String)
         return (String) value;
      return null;
   }
   
   void readDetails() throws AssassinSQLException
   {
      if (dirty()) throw new ObjectModifiedException();

      if (0 == scid) throw new NoDatabaseAssociativityException();
      
      // get all the details of this state change
      String query = "SELECT event, assassin, pluginName, field, value FROM " + DatabaseBackend.getPrefix() + "_statechanges WHERE scid = '"+(new Integer(scid)).toString()+"'";
      ResultSet r = DatabaseBackend.query(query);
      String temps;
      try {
         if (! r.next()) throw new NoDatabaseAssociativityException();
         this.event = Event.getByID(r.getInt("event"));
         this.pluginName = r.getString("pluginName");
         this.field = r.getString("field");
         this.ass = Assassin.getByAid(r.getInt("assassin"));
         temps = r.getString("value");
         r.close();
      } catch (SQLException sqle) {
         throw new AssassinSQLException();
      }

      // check if its a canonical entity, otherwise its a string.
      try {
         this.value = CanonicalEntity.deSerialise(temps);
      } catch (NotSerialisedEntityException NSEe) {
         this.value = temps;
         if (Debug.debug) Debug.print(StateChange.class, "Found a string: "+temps);
         if (Debug.debug) Debug.print(StateChange.class, "Cause: "+NSEe.getMessage());
      }
      
      clean();
   }
   
   void writeDetails() throws AssassinSQLException
   {
      if (Debug.debug) Debug.print(StateChange.class, "Writing out state change "+scid);
      if (!dirty()) return;
      clean();      

      /*
       * 4 cases:
       *
       * 0 == scid && null == event - new, ignore
       * 0 == scid && null != event - new, save
       * 0 != scid && null == event - old, remove
       * 0 != scid && null != event - old, save
       */
      
      // new object - ignore
      if (0 == scid && null == event) {
         return;
      } 
      // new object - save
      else if (0 == scid && null != event) {
         event.writeDetails();
         ass.writeDetails();
         if (value instanceof CanonicalEntity) ((CanonicalEntity) value).writeDetails();
         String sctype = this.getClass().getName();
         if (Debug.debug) Debug.print(StateChange.class, "Details ("+event+","+ass+","+processed+")");
         String query = "INSERT INTO " + DatabaseBackend.getPrefix() + "_statechanges (class,event,assassin,pluginName,field,value, processed) VALUES ('"+sctype+"','"+event.getEid()+"','"+ass.getAid()+"','"+pluginName+"','"+field+"','"+value+"','"+processed.booleanValue()+"')";
         DatabaseBackend.query(query);
         query = "SELECT  LAST_INSERT_ID() FROM " + DatabaseBackend.getPrefix() + "_statechanges";
         // get auto-assigned scid
         ResultSet r;
         try {
            r = DatabaseBackend.query(query);
            if (!r.next()) throw new AssassinSQLException("Oh Ek! That shouldn't happen?!");
            scid = r.getInt(1);
            r.close();
         } catch (SQLException sqle) {
            throw new AssassinSQLException (sqle);
         }
      }

      // existing object - remove
      else if (0 != scid && null == event) {
         String query = "DELETE FROM " + DatabaseBackend.getPrefix() + "_statechanges WHERE scid = '"+scid+"'";
         DatabaseBackend.query(query);
      }

      // existing object - update
      else if (0 != scid && null != event) {
         if (null != this.event)
            this.event.writeDetails();
         String query = "UPDATE " + DatabaseBackend.getPrefix() + "_statechanges SET ";
         if (null != this.event) query += "event = '" + this.event.getEid() + "', ";
         if (null != this.ass) query += "assassin = '" + this.ass.getAid() + "', ";
         if (null != this.pluginName) query += "pluginName = '" + this.pluginName + "', ";
         if (null != this.field) query += "field = '" + this.field + "', ";
         if (null != this.value) query += "value = '" + this.value + "', ";
         if (null != this.processed) query += "processed = '" + this.processed.booleanValue() + "', ";
         query += " scid = '"+scid+"' WHERE scid = '"+scid+"'";
         DatabaseBackend.query(query);
      }
   }
   
   public String toString()
   {
      return "__::" + this.getClass().getName() + "("+(scid)+")";
   }

   public Viewer getViewer()
   {
      return null;
   }

   public String getPrintable() throws AssassinSQLException
   {
      return "State change, set " + getFieldName() + " for " + getAssassin().getPlayer().getName();
   }
}



