/* 
 * CUAG AutoUmpire System
 * 
 * Copyright (C) 2003/2004 Matthew Johnson, Adam Biltcliffe,
 * Michael Cripps, Martin O'Leary, Edward Allcutt and James Osborn
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To Contact the authors, please email assassins@srcf.ucam.org
 *
 */

package org.ucam.srcf.assassins.strawberry;

import java.text.ParseException;

import java.io.IOException;
import java.io.BufferedReader;
import java.io.InputStreamReader;

import java.util.HashMap;
import java.util.Map;
import java.util.Vector;
import java.util.List;
import java.util.Iterator;
import java.util.ArrayList;

import org.ucam.srcf.assassins.AssassinViewer;
import org.ucam.srcf.assassins.AssassinStateViewer;
import org.ucam.srcf.assassins.ConfigProperty;
import org.ucam.srcf.assassins.EventViewer;
import org.ucam.srcf.assassins.Plugin;
import org.ucam.srcf.assassins.PluginConnector;
import org.ucam.srcf.assassins.StateChange;
import org.ucam.srcf.assassins.exceptions.AssassinSQLException;
import org.ucam.srcf.assassins.exceptions.StateBuilderException;
import org.ucam.srcf.assassins.xml.Config;
import org.ucam.srcf.assassins.PluginCommand;
import org.ucam.srcf.assassins.TextQuestion;
import org.ucam.srcf.assassins.Debug;
import org.ucam.srcf.assassins.Death;
import org.ucam.srcf.assassins.Resurrection;
import org.ucam.srcf.assassins.PluginStateChange;
import org.ucam.srcf.assassins.Event;
import org.ucam.srcf.assassins.PlayerViewer;

import org.ucam.srcf.assassins.exceptions.ConfigFileException;

/**
 * Interface for java classes which act as plug-in game modifiers.
 */
public class Strawberry extends Plugin
{
   public static String getDescription() { return "Scoring plugin"; };
	int defaultStartingBaseScore;
	
	Scorer AbaseScore;
	Scorer VbaseScore;
	Scorer AScoreMod;
	Scorer VScoreMod;

	
   /**
    * A reference back to the interface with the base system
    */
   protected PluginConnector vanilla;
   
   /**
    * Sets the canonical string representing the Plugin
    * @param s The string representing the Plugin
    */
   protected void setPluginName(String s) {
		super.setPluginName(s);
		
		defaultStartingBaseScore = (Integer.valueOf(Config.getConfig("strawberry", "default-starting-score"))).intValue();

		try {
			AbaseScore=new Scorer(Config.getConfig("strawberry", "AbaseScoreEqn"),0);
			VbaseScore=new Scorer(Config.getConfig("strawberry", "VbaseScoreEqn"),0);
			AScoreMod=new Scorer(Config.getConfig("strawberry", "AScoreModEqn"),0);
			VScoreMod=new Scorer(Config.getConfig("strawberry", "VScoreModEqn"),0);
		}
		catch (ConfigFileException cfex) {
			Debug.print(this, "Dodgy config file");
			cfex.printStackTrace();
		}
	}
   
   /**
    * Called whenever the plugin is required to update or sanity-check the
    * current gamestate.
    */
   protected Vector updateState(Map state, Vector qs, EventViewer et) 
		throws AssassinSQLException, ParseException 
	{ 
		/* We proceed to go through the entire system and check the scores 
		 * assigned at each "death" event. This officially sucks, as we have
		 * to build our own state for aliveness
		 */
		/* I have been assured that noone would be silly enough to think
		 * that there should eb intermediate scores worked out during an
		 * event, and that noone will put a chain of deaths
		 * A->B->C and expect A to get some of B's points for killing
		 * C from this event.
		 */

		/* Set up a var tracker to track stuff */
		VarTracker vt = new VarTracker(state, getPluginName());

		/* Get all events in chronological order */
		List events = Event.getAllEventViewersAtTime(vanilla.getNowTime());
		Iterator i = events.iterator();
		while(i.hasNext()) {
			EventViewer e = (EventViewer) i.next();
			if(!e.isActive())
				continue;
			List sclist = e.getStateChanges();
			Iterator j = sclist.iterator();

			/*Set up a map from assassins to *modified* scores*/
			Map scores = vt.refreshScores(e);
			
			/* List of score StateChanges so we can check if they need updating at 
			 * the end of processing an event
			 */
			List scoreChanges = new ArrayList();

			/*Now iterate through the statechanges*/
			while(j.hasNext()) {
				StateChange sc = (StateChange) j.next();
				AssassinViewer a = sc.getAssassinViewer();
				String pname = sc.getPluginName();
				String field = sc.getFieldName();
				Object value = sc.getValue();
				if(sc instanceof Death) {
					/* Score it and notify the VarTracker. */
					scoreDeath((Death) sc, vt, scores);
					vt.trackDeath((Death) sc);
					continue;
				} 
				if(sc instanceof Resurrection) {
					vt.setProperty(a,"alive","true");
					continue;
				} 
				if(!pname.equals(this.getPluginName()))
					continue;
				/* Otherwise we have a sc applicable to this plugin, */

				/* If its a score change we deal with it later */
				if(	field.equals("BaseScore") ||
						field.equals("PostGameMod") ||
						field.equals("FinalScore")
						) {
					scoreChanges.add(sc);
					continue;
				}

				/* If its an innocent death we deal with it */
				if(field.equals("Innocents")) {
					int numKilled = Integer.parseInt((String) value);
					for(int innocent=0; innocent<numKilled; innocent++) {
						scoreInnocentDeath(a, vt, scores);
						vt.trackInnocent(a);
					}
				}

				/*If its a bonus we add it to the scores*/
				if(field.equals("BaseBonus")) {
					AssassinScore as=((AssassinScore) scores.get(a));
					if(as==null) 
						scores.put(a, as=new AssassinScore(a, this.getPluginName(), vt));
					as.addBase(Integer.parseInt((String) value));
					continue;
				}
				if(field.equals("ModBonus")) {
					AssassinScore as=((AssassinScore) scores.get(a));
					if(as==null)
						scores.put(a, as=new AssassinScore(a, this.getPluginName(), vt));
					as.addPost(Integer.parseInt((String) value));
					continue;
				}
				
				/* otherwise we just apply it */
				vt.setProperty(a, field, (String) value);
				if(Debug.debug) Debug.print(this, "Applying unknown Strawberry State Change (" + field + ", " + value + ")");
			} /*End of states loop */

			/* Now we go through and check if the scores are right or not */
			/* First leave anything which is correct alone */
			j = scoreChanges.iterator();
			while(j.hasNext()) {
				StateChange sc = (StateChange) j.next();
				AssassinViewer a = sc.getAssassinViewer();
				String field = sc.getFieldName();
				int value = Integer.parseInt((String) sc.getValue());
				if(scores.get(a) == null) {
					/*merley update our intenal state*/
					vt.setProperty(a, field, String.valueOf(value));
					continue;
				}
				AssassinScore cur = (AssassinScore) scores.get(a);
				if(field.equals("BaseScore")) {
					if(cur.getBaseScore() != value)
						vanilla.removeStateChange(e, sc);
					else
						cur.baseClean();
					continue;
				}
				if(field.equals("PostGameMod")) {
					if(cur.getPostGameMod() != value)
						vanilla.removeStateChange(e, sc);
					else
						cur.postClean();
					continue;
				}
				if(field.equals("FinalScore")) {
					if(cur.getFinalScore() != value)
						vanilla.removeStateChange(e, sc);
					else
						cur.finalClean();
					continue;
				}
			}
			/* Now we add any dirty score changes to the event */
			j = scores.keySet().iterator();
			while(j.hasNext()) {
				AssassinScore cur=(AssassinScore) scores.get(j.next());
				AssassinViewer a = cur.getAssassinViewer();
				/*write out if they need it*/
				if(cur.isBaseDirty())
					vanilla.addStateChange(e, cur.getBaseSC(this));
				if(cur.isPostDirty())
					vanilla.addStateChange(e, cur.getPostSC(this));
				if(cur.isFinalDirty())
					vanilla.addStateChange(e, cur.getFinalSC(this));
				/*and update our internal state*/
				vt.setIntProperty(a, "BaseScore", cur.getBaseScore());
				vt.setIntProperty(a, "PostGameMod", cur.getPostGameMod());
				vt.setIntProperty(a, "FinalScore", cur.getFinalScore());
			}
				
		} /*end of events loop */
				
		return null; 
	}

	private void scoreInnocentDeath(AssassinViewer a, VarTracker vt, Map scores) {
		int newAbaseScore = (int) Math.round(AbaseScore.execute(a, null, vt));
		int newApostGameMod = (int) Math.round(AScoreMod.execute(a, null, vt));
		
		AssassinScore aScore = (AssassinScore) scores.get(a);

		if(aScore==null)
			scores.put(a, aScore=new AssassinScore(a, this.getPluginName(), vt));
		
		aScore.setBaseScore(newAbaseScore);
		aScore.setPostGameMod(newApostGameMod);
	}
	
	private void scoreDeath(Death myDeath, VarTracker vt, Map scores) {
      /* calculate the new scores */
      try {
         int newAbaseScore = (int) Math.round(AbaseScore.execute(myDeath
               .getKiller(), myDeath.getVictim(), vt));
         int newVbaseScore = (int) Math.round(VbaseScore.execute(myDeath
               .getKiller(), myDeath.getVictim(), vt));
         int newApostGameMod = (int) Math.round(AScoreMod.execute(myDeath
               .getKiller(), myDeath.getVictim(), vt));
         int newVpostGameMod = (int) Math.round(VScoreMod.execute(myDeath
               .getKiller(), myDeath.getVictim(), vt));

         /* get the current scores if they exist, otherwise make them */
         AssassinScore aScore = (AssassinScore) scores.get(myDeath.getKiller());
         AssassinScore vScore = (AssassinScore) scores.get(myDeath.getVictim());
         if (aScore == null)
               scores.put(myDeath.getKiller(), aScore = new AssassinScore(
                     myDeath.getKiller(), this.getPluginName(), vt));
         if (vScore == null)
               scores.put(myDeath.getVictim(), vScore = new AssassinScore(
                     myDeath.getVictim(), this.getPluginName(), vt));
         /* now set the scores correctly */
         aScore.setBaseScore(newAbaseScore);
         aScore.setPostGameMod(newApostGameMod);

         vScore.setBaseScore(newVbaseScore);
         vScore.setPostGameMod(newVpostGameMod);
      } catch (AssassinSQLException asqle) {
         return;
      }
   }

   /**
    * Called when an event is being added to get any queries the plugin
    * has for the user. The plugin should check all the event details and if it
    * needs more information it should ask relevent questions. The 
    * returnQuestions function is passed the same Vector with the answers 
    * filled out.
    * @param e The Event which is being added
    * @return A Vector of Questions
    * @see returnQuestions
    */
   protected Vector getQuestions(EventViewer e, Map state) 
		throws AssassinSQLException 
	{
		Vector qs=new Vector();

		Iterator i = e.getParticipants().values().iterator();
		while(i.hasNext()) {
			AssassinViewer a= (AssassinViewer) i.next();
			PlayerViewer p = a.getPlayerViewer();
			qs.add(new TextQuestion("Give " + p.getName() + 
						" a transfferable bonus of:", "0"));
			qs.add(new TextQuestion("Give " + p.getName() + 
						" a non-transfferable bonus of:", "0"));
			qs.add(new TextQuestion("Number of innocents " + p.getName() + 
						" killed:", "0"));
		}
				
		return qs;
	}
   
   /**
    * Returns the Questions with the answers filled out
    *
    * @param e The Event being added
    * @param qs The Questions which were asked
    */
   protected void returnQuestions(EventViewer e, Map state, Vector qs) 
		throws AssassinSQLException 
	{
		Iterator i = e.getParticipants().values().iterator();
		while(i.hasNext()) {
			AssassinViewer a= (AssassinViewer) i.next();
			PlayerViewer p = a.getPlayerViewer();
			String baseScoreBonus = ((TextQuestion) qs.get(0)).getAnswer();
			String postGameBonus = ((TextQuestion) qs.get(1)).getAnswer();
			String innocents = ((TextQuestion) qs.get(2)).getAnswer();
			vanilla.addStateChange(e, new PluginStateChange(a, this, "BaseBonus", baseScoreBonus));
			vanilla.addStateChange(e, new PluginStateChange(a, this, "ModBonus", postGameBonus));
			if(!innocents.equals("0"))
				vanilla.addStateChange(e, new PluginStateChange(a, this, "Inoccents", innocents));
		}
		
	}

   /**
    * Called when an assassin is being added to get any queries the plugin
    * has for the user. The plugin should check all the assassin details and if it
    * needs more information it should ask relevent questions. The 
    * returnQuestions function is passed the same Vector with the answers 
    * filled out.
    * @param a The Assassin which is being added
    * @return A Vector of Questions
    * @see returnQuestions
    */
   protected Vector getQuestions(AssassinViewer a, Map state) 
		throws AssassinSQLException 
	{
		Vector v=new Vector();
		
		v.add(new TextQuestion("Starting Base Score", ""+defaultStartingBaseScore));
		v.add(new TextQuestion("Starting Non-transferable Score Modifier", "0"));

		return v;
	}
   
   /**
    * Returns the Questions with the answers filled out
    *
    * @param a The Assassin being added
    * @param e An Event to add StateChanges to
    * @param qs The Questions which were asked
    */
   protected void returnQuestions(AssassinViewer a, EventViewer e, Map state, Vector qs) 
		throws AssassinSQLException 
	{
		/* We expect the following questions:
		 * 0	-	Starting Base Score			(BaseScore)
		 * 1	-	Starting Score Modifier		(PostGameMod)
		 */
		
		int baseScore = Integer.parseInt(((TextQuestion) qs.get(0)).getAnswer());
		int postGameMod = Integer.parseInt(((TextQuestion) qs.get(1)).getAnswer());

		vanilla.addStateChange(e, new PluginStateChange(a, this, "BaseScore", String.valueOf(baseScore)));
		vanilla.addStateChange(e, new PluginStateChange(a, this, "PostGameMod", String.valueOf(postGameMod)));
	}

   /**
    * Returns a Map of new commands to add to the CLI
    *
    * @return A Map of String =&gt; Command
    */
   public Map getCommands() { 
		Map m = new HashMap();
		final Strawberry st = this;
		
		m.put("score-assassin",
				new PluginCommand() {
					public void pluginCommand(Map params) throws AssassinSQLException {
						st.scoreAssassin(params);
					}
				}
		);
		
		return m; 
	}

   /**
    * Returns a Map of new commands to add to the CGI
    *
    * @return A Map of String =&gt; CGICommand
    */
   public Map getCGICommands() { 
		return new HashMap(); 
	}

	private AssassinViewer selectAssassin(Map state) 
		throws IOException, AssassinSQLException
	{
		List names=new ArrayList();

		Iterator i=state.keySet().iterator();
		while(i.hasNext())
			names.add(((AssassinViewer) i.next()).getPlayerViewer().getName());
		
		int index = askOption("Select Assassin:", names, 0) - 1;
		String name = (String) names.get(index);
		
		i=state.keySet().iterator();
		AssassinViewer current=null;
		while(i.hasNext())
			if((current=((AssassinViewer) i.next())).getPlayerViewer().getName().equals(name))
				return current;
					
		return null;
	}

	private int askOption(String prompt, List options, int def) 
		throws IOException
	{
		boolean notDone=true;
		BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
		int ans=0;
		
		while(notDone) {
			System.out.println(prompt);
			Iterator i=options.iterator();
			int j=0;
			while(i.hasNext())
				System.out.println((j++) + ". " + ((String) i.next())); 
			if(def!=0)
				System.out.print("Enter a number ["+def+"]:");
			else
				System.out.print("Enter a number:");
			String selection=in.readLine();
			if(selection.equals(""))
				return def;
			try {
				if(	(ans=Integer.parseInt(selection)) > 0 &&
						Integer.parseInt(selection) < options.size()		)
					notDone=false;
				else
					notDone=true;
			} catch (NumberFormatException ex) {
				notDone=true;
			}
		}
		return ans;
	}

	public void scoreAssassin(Map params) 
		throws AssassinSQLException 
	{
      // TODO: should be able to only build state for one assassin here?
      Map state = vanilla.getState();
		try {
			updateState(state, null, null);
		}
		catch (ParseException pe) {
			System.out.println("Failed to build Stawberry state, exiting.");
		}
		
		AssassinViewer assassin;
		try {
			assassin = selectAssassin(state);
		} catch (IOException IOe) {
			System.out.println("Error while readin stdin, terminating.");
			return;
		}
		AssassinStateViewer aState = (AssassinStateViewer) state.get(assassin);
		
		System.out.println("Scoring assassin " + aState.getPseudonyms().get(aState.getPseudonyms().size()));
		
		int baseScore = aState.getIntProperty(this.getPluginName(), "BaseScore");
		int postGameMod = aState.getIntProperty(this.getPluginName(), "PostGameMod");
		int finalScore = aState.getIntProperty(this.getPluginName(), "FinalScore");
		
		System.out.println("Final Score = Base Score + Post Game Modifier");
		System.out.println(finalScore + " = " + baseScore + " + " + postGameMod);
	}
   public  static Vector getConfigProperties()
   {
      Vector v = new Vector();
      v.add ( new ConfigProperty ("default-starting-score", "", ""));
      v.add ( new ConfigProperty ("AbaseScoreEqn", "", ""));
      v.add ( new ConfigProperty ("VbaseScoreEqn", "", ""));
      v.add ( new ConfigProperty ("AScoreModEqn", "", ""));
      v.add ( new ConfigProperty ("VScoreModEqn", "", ""));
   }

}


